import email
from email.header import decode_header
import getpass
from halo import Halo
import imaplib
import json
from os import path
import pandas as pd
import re
from tqdm import tqdm


class ImapConnection:
    """
    Establishes a IMAP connection to an email account
    """
    def __init__(self):
        """
        Asks the credentials on the terminal, connects and gets the account's data
        If the email is not on gmail or outlook, the host and port are requested as well
        Gets default maximum number of loaded emails from settings.json
        """

        self.__get_settings()
        self.__creds = self.__get_creds()

        if self.__creds['email_user'][-10:] == '@gmail.com':
            host = 'imap.gmail.com'
            port = 993
        elif self.__creds['email_user'][-12:] == '@outlook.com':
            host = 'imap-mail.outlook.com'
            port = 993
        else:
            host = input('Enter imap host: ')
            port = input('Enter imap port: ')

        spinner = Halo(text='Connecting', spinner='dots')
        spinner.start()

        self.account = imaplib.IMAP4_SSL(host, port)
        self.account.login(self.__creds['email_user'], self.__creds['email_pass'])

        spinner.stop()
        print(f'\nConnected to mail account')


    def __get_creds(self):
        """
        Requests the credentials in the terminal

        Returns:
            dict: credentials dictionary
        """
        email_user = input('User: ')
        email_pass = getpass.getpass()

        creds = {
            'email_user': email_user,
            'email_pass': email_pass
        }

        return creds


    def __get_settings(self):
        """
        Gets the settings from the settings.json file

        Raises:
            FileNotFoundError: if the settings.json file is not found in the parent path of the program
        """
        current_dir = path.dirname(path.abspath(__file__))
        settings_path = current_dir + '/../settings.json'
        if path.exists(settings_path):
            with open(settings_path, 'r') as f:
                json_settings = f.read()
            self.__settings = json.loads(json_settings)
        else:
            raise FileNotFoundError("File settings.json not found")


    def get_emails_from_folder(self, folder):
        """
        Fetches emails from a folder
        The maximum number of retrieved emails at program start comes from the file settings.json
        but it can be changed from within the user interface menu with option 's'

        Args:
            folder (str): mailbox folder like 'inbox"

        Returns:
            dataframe: dataframe containing all emails
        """
        self.max_loaded_mails = self.__settings['max_default_loaded_mails']

        print(f'Fetching {self.max_loaded_mails} emails...\n')
        msgs = self.__get_messages_from_folder(folder)
        df = self.__msgs_to_df(msgs)

        return df


    def __get_messages_from_folder(self, folder):
        """
        Fetches emails from a given email folder like inbox

        Args:
            folder (str): mailbox folder like 'inbox'

        Returns:
            list: list of tuples containing a message ID and a message object
        """
        self.account.select(folder)
        data = self.account.search(None, 'ALL')

        mail_ids = data[1]
        id_list = mail_ids[0].split()

        print(f'\nYou have {len(id_list)} emails in your inbox\n')

        first_email_id = int(id_list[-1])
        last_email_id = int(id_list[-self.max_loaded_mails-1])

        msgs = []

        for id in tqdm(range(first_email_id, last_email_id, -1)):

            data = self.account.fetch(str(id), '(RFC822)' )

            for response_part in data:
                arr = response_part[0]
                if isinstance(arr, tuple):
                    try:
                        msg = email.message_from_string(str(arr[1], 'utf-8'))
                    except:
                        msg = email.message_from_string(str(arr[1], 'unicode_escape'))
                    msgs.append((str(id), msg)) # will generate a list of tuples id, message

        return msgs


    def __msgs_to_df(self, msgs):
        """
        Transforms list of tuples containing message ID and message object into a pandas dataframe

        Args:
            msgs (list): list of tuples containing a message ID and a message object

        Returns:
            dataframe: dataframe containing all emails
        """
        data_list = []

        for msg in msgs:
            email = msg[1] # msg(0) is the ID
            subject = self.__get_header_component(email, 'subject')
            From = self.__get_header_component(email, 'From').split('<')[0].split(',')[0].replace('"', '').strip()
            date = self.__get_header_component(email, 'date')
            body = self.__get_body(email)
            if self.__settings['remove_urls']:
                body = self.__remove_urls(body)
            else:
                pass
            uid = msg[0]
            data_list.append([subject, From, date, body, uid])

        col_list = ['Subject', 'Sender', 'Date', 'Body', 'Uid']
        df = pd.DataFrame(data_list, columns=col_list)
        df['Date'] = pd.to_datetime(df['Date'], infer_datetime_format=True, format='%d/%b/%Y, %H:%M:%S')

        return df


    def __get_header_component(self, msg, component):
        """
        Gets the header component for a message

        Args:
            msg (obj): email message object
            component (str): header component

        Returns:
            str: component content
        """
        try:
            component, encoding = decode_header(msg[component])[0]
            if isinstance(component, bytes):
                # if it's a bytes, decode to str
                if encoding:
                    component = component.decode(encoding)
                else:
                    component = component.decode()
        except:
            component=''
        return component


    def __remove_urls(self, text):
        """
        Removes urls from a string

        Args:
            text (str): any given string

        Returns:
            string: string without url links
        """
        return re.sub(r'https?://\S+', '', text)


    def __get_body(self, msg):
        """
        Gets the body from a list an email object

        Args:
            msg (obj): email object

        Returns:
            str: email body
        """
        body = ''
        if msg.is_multipart():
            for part in msg.walk():
                if part.get_content_type() == "text/plain":
                    charset = part.get_content_charset()
                    body_part = part.get_payload(decode=True)
                    try:
                        body += body_part.decode(charset)
                    except:
                        body += body_part.decode('unicode_escape')
                    return body
        else:
            body_part = msg.get_payload()
            if isinstance(body_part, bytes):
                body = body_part.decode()
            else:
                body = body_part

        return body



from datetime import datetime
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from tabulate import tabulate
from tqdm import tqdm
import webbrowser

import warnings
warnings.filterwarnings('ignore')


class ui:
    def __init__(self, msgs_df, imap_connection):
        self.imap_connection = imap_connection
        self.account = imap_connection.account
        self.msgs_df = msgs_df
        self.df_to_display = self.msgs_df[['Sender', 'Subject', 'Date']]
        self.__option = ''

        # Print last 10 received mails and total fetched messages
        self.__display_df(10)

        # Render user interface
        while True:

            # Request user option
            self.__option = input('\nOption: ')

            # Quit
            if self.__option == 'q':
                self.__quit()
                break

            # Delete all messages from selected sender
            elif self.__option == 'd':
                self.__delete_by_subject()

            # Delete selected email by dataframe index number
            elif self.__option[0] == 'd' and self.__option[1:].isnumeric():
                self.__delete_by_index()

            # Filter
            elif self.__option == 'f':
                self.__filter_by_sender()

            # Graph top 10 senders
            elif self.__option == 'g':
                self.__graph_top_senders()

            # Help
            elif self.__option == 'h' or self.__option == 'help':
                self.__help()

            # List top 20 mails
            elif self.__option == 'l':
                self.__display_df(20)

            # List selected amount of emails
            elif self.__option[0] == 'l' and self.__option[1:].isnumeric():
                no_of_lines = int(self.__option[1:])
                self.__display_df(no_of_lines)

            # Reload mails from server
            elif self.__option == 'r':
                self. __reload_mails()

            # Set maximum number of loaded mails
            elif self.__option == 's':
                self.__set_max_mails()

            # Show message content by index number
            elif self.__option.isnumeric():
                msg_no = int(self.__option)
                subject = self.msgs_df.loc[msg_no, 'Subject']
                From = self.msgs_df.loc[msg_no, 'Sender']
                body = self.msgs_df.loc[msg_no, 'Body']
                print(f'\n--- Subject: {subject} ')
                print(f'\n--- From: {From} ')
                print(f'\n{body}\n')

            # Show message content by index number as html in the browser
            elif (self.__option.split(' ')[0].isnumeric() and self.__option.split(' ')[1] == 'as' and self.__option.split(' ')[2] == 'html'):
                msg_no = int(self.__option.split(' ')[0])
                # subject = self.msgs_df.loc[msg_no, 'Subject']
                # From = self.msgs_df.loc[msg_no, 'Sender']
                body = self.msgs_df.loc[msg_no, 'Body']
                temp_html = 'temp' + datetime.now().time().strftime('%H%M%S') + '.html'
                with open (temp_html, 'w') as f:
                    f.write(body)
                webbrowser.open(temp_html)
            else:
                print('Invalid option. Try again.')


    def __quit(self):
        # close the connection and logout
        self.account.close()
        self.account.logout()


    def __delete_by_subject(self):
        selected_sender = input('\nEnter the exact sender from whom you would like to delete ALL emails: ')
        yes_no_delete = input('Are you sure you would like to delete this message (y/n)? ')
        if yes_no_delete == 'y':
            print(f'Selected sender: {selected_sender}')
            filtered_df = self.msgs_df[self.msgs_df.Sender == selected_sender].reset_index()
            rows_filtered = filtered_df.shape[0]
            print(f'\n{rows_filtered} message(s) successfully deleted.\n')
            for i in tqdm(range(0, rows_filtered)):
                self.account.store(filtered_df.loc[i, 'Uid'], "+FLAGS", "\\Deleted")
                self.account.expunge()
            self.df_to_display = self.df_to_display[self.df_to_display.Sender != selected_sender]
            
        else:
            print('Ok, messages were not deleted')


    def __delete_by_index(self):
        msg_no = int(self.__option[1:])
        yes_no_delete = input('Are you sure you would like to delete this message (y/n)? ')
        if yes_no_delete == 'y':
            subject = self.msgs_df.loc[msg_no, 'Subject']
            self.account.store(self.msgs_df.loc[msg_no, 'Uid'], '+FLAGS', '\\Deleted')
            self.account.expunge()
            self.df_to_display.drop(msg_no, inplace=True)
            print(f'\n--- Subject: {subject} ')
            print('\nMessage successfully deleted.\n')
        elif yes_no_delete == 'n':
            print('Ok, message was not deleted')
        else:
            print('Invalid answer.')


    def __filter_by_sender(self):
        by_sender = input('Enter sender to filter by: ')
        df_filtered = self.df_to_display[self.df_to_display.Sender == by_sender]
        pd.set_option('display.max_colwidth', None)
        print(tabulate(df_filtered, headers = 'keys', tablefmt = 'psql'))


    def __graph_top_senders(self):
        sender_count  = self.msgs_df['Sender'].value_counts()
        sender_count = sender_count[:10,]
        plt.figure(figsize=(10,5))
        sns.barplot(sender_count.index, sender_count.values, alpha=0.8)
        plt.title('Top 10 senders to my email', fontsize=14)
        plt.ylabel('Number of Occurrences', fontsize=10)
        plt.xticks(fontsize=8)
        plt.xlabel('Senders', fontsize=10)
        plt.show()


    @staticmethod
    def __help():
        print('\n')
        print(f'd - Delete by sender')
        print(f'd[index] - Delete by [index]')
        print(f'f - Filter by sender')
        print(f'g - Graph top 10 senders')
        print(f'h - Help manual')
        print(f'l - List last 20 mails')
        print(f'l[number] - List last [number] mails')
        print(f'r - Reload mails from server')
        print(f's - Set number of loaded mails from server')
        print(f'[number] - Show selected mail on console')
        print(f'[number] as html - Show selected mail on browser')
        print('\n')

    def __reload_mails(self):
        print('Refreshing emails...\n')
        self.msgs_df = self.imap_connection.get_emails_from_folder('inbox')
        self.df_to_display = self.msgs_df[['Sender', 'Subject', 'Date']]


    def __set_max_mails(self):
        new_max_loaded_mails = input('Set maximum number of loaded emails (press r to reload after this): ')
        if not new_max_loaded_mails.isnumeric():
            print('Only numbers accepted. Try again.')
        else:
            self.imap_connection.max_loaded_mails = int(new_max_loaded_mails)


    def __display_df(self, no_of_lines):
        print(f'\n--- Your last {no_of_lines} received emails ---\n')
        pd.options.display.max_columns = None
        pd.options.display.max_colwidth = None
        self.df_to_display["Subject"] = self.df_to_display["Subject"].str.wrap(50)
        print(tabulate(self.df_to_display.head(no_of_lines), headers = 'keys', tablefmt = 'psql'))
        print(f'Total fetched messages: {self.msgs_df.shape[0]}')



import pathlib
from setuptools import setup

# The directory containing this file
HOME_DIR = pathlib.Path(__file__).parent

# The text of the README file
README = (HOME_DIR / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="pymailorganizer",
    version="0.6.0",
    description="Python command line mail organizer",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/santosgo/pymailorganizer",
    author="Gonçalo Brandão Coelho dos Santos",
    author_email="goncalo.a.santos@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["pymailorganizer"],
    include_package_data=True,
    install_requires=[
        "halo", 
        "matplotlib", 
        "ntlm-auth",
        "numpy",
        "pandas",
        "pytest",
        "requests",
        "scipy",
        "seaborn",
        "tabulate",
        "tqdm",
        "urllib3",
    ],
    entry_points={
        "console_scripts": [
            "pymailorganizer = pymailorganizer.main:shell",
        ]
    },
)
